import io
from pathlib import Path
from types import GeneratorType
from typing import Any, Callable


try:
    import numpy as np  # type: ignore

    has_numpy = True
except ImportError:
    has_numpy = False


def encode_json(obj: Any, upload_file: Callable[[io.IOBase], str]) -> Any:
    """
    Returns a JSON-compatible version of the object. Effectively the same thing as cog.json.encode_json.
    """
    if isinstance(obj, dict):
        return {key: encode_json(value, upload_file) for key, value in obj.items()}
    if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
        return [encode_json(value, upload_file) for value in obj]
    if isinstance(obj, Path):
        with obj.open("rb") as f:
            return upload_file(f)
    if isinstance(obj, io.IOBase):
        return upload_file(obj)
    if has_numpy:
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
    return obj



from replicate.base_model import BaseModel


class Collection:
    """
    A base class for representing all objects of a particular type on the
    server.
    """

    model: BaseModel = None

    def __init__(self, client=None):
        self._client = client

    def list(self):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError

    def create(self, attrs=None):
        raise NotImplementedError

    def prepare_model(self, attrs):
        """
        Create a model from a set of attributes.
        """
        if isinstance(attrs, BaseModel):
            attrs._client = self._client
            attrs._collection = self
            return attrs
        elif isinstance(attrs, dict):
            model = self.model(**attrs)
            model._client = self._client
            model._collection = self
            return model
        else:
            raise Exception("Can't create %s from %s" % (self.model.__name__, attrs))



import datetime
from typing import Any, Iterator, List, Union

from replicate.base_model import BaseModel
from replicate.collection import Collection
from replicate.exceptions import ModelError


class Version(BaseModel):
    id: str
    created_at: datetime.datetime
    cog_version: str
    openapi_schema: Any

    def predict(self, **kwargs) -> Union[Any, Iterator[Any]]:
        # TODO: support args
        prediction = self._client.predictions.create(version=self, input=kwargs)
        # Return an iterator of the output
        # FIXME: might just be a list, not an iterator. I wonder if we should differentiate?
        if (
            self.openapi_schema["components"]["schemas"]["Output"].get("type")
            == "array"
        ):
            return prediction.output_iterator()

        prediction.wait()
        if prediction.status == "failed":
            raise ModelError(prediction.error)
        return prediction.output


class VersionCollection(Collection):
    model = Version

    def __init__(self, client, model):
        super().__init__(client=client)
        self._model = model

    # doesn't exist yet
    def get(self, id: str) -> Version:
        """
        Get a specific version.
        """
        resp = self._client._request(
            "GET", f"/v1/models/{self._model.username}/{self._model.name}/versions/{id}"
        )
        return self.prepare_model(resp.json())

    def list(self) -> List[Version]:
        """
        Return a list of all versions for a model.
        """
        resp = self._client._request(
            "GET", f"/v1/models/{self._model.username}/{self._model.name}/versions"
        )
        return [self.prepare_model(obj) for obj in resp.json()["results"]]



import base64
import io
import mimetypes
import os

import requests


def upload_file(fh: io.IOBase, output_file_prefix: str = None) -> str:
    """
    Lifted straight from cog.files
    """
    fh.seek(0)

    if output_file_prefix is not None:
        name = getattr(fh, "name", "output")
        url = output_file_prefix + os.path.basename(name)
        resp = requests.put(url, files={"file": fh})
        resp.raise_for_status()
        return url

    b = fh.read()
    # The file handle is strings, not bytes
    if isinstance(b, str):
        b = b.encode("utf-8")
    encoded_body = base64.b64encode(b)
    if getattr(fh, "name", None):
        # despite doing a getattr check here, mypy complains that io.IOBase has no attribute name
        mime_type = mimetypes.guess_type(fh.name)[0]  # type: ignore
    else:
        mime_type = "application/octet-stream"
    s = encoded_body.decode("utf-8")
    return f"data:{mime_type};base64,{s}"



import time
from typing import Any, Dict, Iterator, List, Optional

from replicate.base_model import BaseModel
from replicate.collection import Collection
from replicate.exceptions import ModelError, ReplicateException
from replicate.files import upload_file
from replicate.json import encode_json
from replicate.version import Version


class Prediction(BaseModel):
    id: str
    error: Optional[str]
    input: Optional[Dict[str, Any]]
    logs: Optional[str]
    output: Optional[Any]
    status: str
    version: Optional[Version]

    def wait(self):
        """Wait for prediction to finish."""
        while self.status not in ["succeeded", "failed", "canceled"]:
            time.sleep(0.1)
            self.reload()

    def output_iterator(self) -> Iterator[Any]:
        # TODO: check output is list
        previous_output = self.output or []
        while self.status not in ["succeeded", "failed"]:
            output = self.output or []
            new_output = output[len(previous_output) :]
            for output in new_output:
                yield output
            previous_output = output
            time.sleep(0.1)
            self.reload()

        if self.status == "failed":
            raise ModelError(self.error)

        output = self.output or []
        new_output = output[len(previous_output) :]
        for output in new_output:
            yield output

    def cancel(self):
        """Cancel a currently running prediction"""
        self._client._request("POST", f"/v1/predictions/{self.id}/cancel")


class PredictionCollection(Collection):
    model = Prediction

    def create(self, version: Version, input: Dict[str, Any]) -> Prediction:
        input = encode_json(input, upload_file=upload_file)
        resp = self._client._request(
            "POST", "/v1/predictions", json={"version": version.id, "input": input}
        )
        obj = resp.json()
        obj["version"] = version
        return self.prepare_model(obj)

    def get(self, id: str) -> Prediction:
        resp = self._client._request("GET", f"/v1/predictions/{id}")
        obj = resp.json()
        # HACK: resolve this? make it lazy somehow?
        del obj["version"]
        return self.prepare_model(obj)

    def list(self) -> List[Prediction]:
        resp = self._client._request("GET", f"/v1/predictions")
        # TODO: paginate
        predictions = resp.json()["results"]
        for prediction in predictions:
            # HACK: resolve this? make it lazy somehow?
            del prediction["version"]
        return [self.prepare_model(obj) for obj in predictions]



import os
from json import JSONDecodeError

import requests
from requests.adapters import HTTPAdapter, Retry

from replicate.__about__ import __version__
from replicate.exceptions import ReplicateError
from replicate.model import ModelCollection
from replicate.prediction import PredictionCollection


class Client:
    def __init__(self, api_token=None) -> None:
        super().__init__()
        # Client is instantiated at import time, so do as little as possible.
        # This includes resolving environment variables -- they might be set programmatically.
        self.api_token = api_token
        self.base_url = os.environ.get(
            "REPLICATE_API_BASE_URL", "https://api.replicate.com"
        )

        # TODO: make thread safe
        self.session = requests.Session()

        # Gracefully retry requests
        # This is primarily for when iterating through predict(), where if an exception is thrown, the client
        # has no way of restarting the iterator.
        # We might just want to enable retry logic for iterators, but for now this is a blunt instrument to
        # make this reliable.
        retries = Retry(
            total=5,
            backoff_factor=2,
            # Only retry on GET so we don't unintionally mutute data
            method_whitelist=["GET"],
            # https://support.cloudflare.com/hc/en-us/articles/115003011431-Troubleshooting-Cloudflare-5XX-errors
            status_forcelist=[500, 502, 503, 504, 520, 521, 522, 523, 524, 526, 527],
        )
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

    def _request(self, method: str, path: str, **kwargs):
        # from requests.Session
        if method in ["GET", "OPTIONS"]:
            kwargs.setdefault("allow_redirects", True)
        if method in ["HEAD"]:
            kwargs.setdefault("allow_redirects", False)
        kwargs.setdefault("headers", {})
        kwargs["headers"].update(self._headers())
        resp = self.session.request(method, self.base_url + path, **kwargs)
        if 400 <= resp.status_code < 600:
            try:
                raise ReplicateError(resp.json()["detail"])
            except (JSONDecodeError, KeyError):
                pass
            raise ReplicateError(f"HTTP error: {resp.status_code, resp.reason}")
        return resp

    def _headers(self):
        return {
            "Authorization": f"Token {self._api_token()}",
            "User-Agent": f"replicate-python@{__version__}",
        }

    def _api_token(self):
        token = self.api_token
        # Evaluate lazily in case environment variable is set with dotenv, or something
        if token is None:
            token = os.environ.get("REPLICATE_API_TOKEN")
        if not token:
            raise ReplicateError(
                """No API token provided. You need to set the REPLICATE_API_TOKEN environment variable or create a client with `replicate.Client(api_token=...)`.

You can find your API key on https://replicate.com"""
            )
        return token

    @property
    def models(self) -> ModelCollection:
        return ModelCollection(client=self)

    @property
    def predictions(self) -> PredictionCollection:
        return PredictionCollection(client=self)



class ReplicateException(Exception):
    pass


class ModelError(ReplicateException):
    """An error from user's code in a model."""


class ReplicateError(ReplicateException):
    """An error from Replicate."""



from typing import ForwardRef
import pydantic


Client = ForwardRef("Client")
Collection = ForwardRef("Collection")


class BaseModel(pydantic.BaseModel):
    """
    A base class for representing a single object on the server.
    """

    _client: Client = pydantic.PrivateAttr()
    _collection: Collection = pydantic.PrivateAttr()

    def reload(self):
        """
        Load this object from the server again.
        """
        new_model = self._collection.get(self.id)
        for k, v in new_model.dict().items():
            setattr(self, k, v)



from typing import List

from replicate.base_model import BaseModel
from replicate.collection import Collection
from replicate.exceptions import ReplicateException
from replicate.version import VersionCollection


class Model(BaseModel):
    username: str
    name: str

    def predict(self, *args, **kwargs):
        versions = self.versions.list()
        if not versions:
            raise ReplicateException(
                "No versions found for model %s/%s" % (self.username, self.name)
            )
        latest_version = versions[0]
        return latest_version.predict(*args, **kwargs)

    @property
    def versions(self):
        return VersionCollection(client=self._client, model=self)


class ModelCollection(Collection):
    model = Model

    def get(self, name: str) -> Model:
        # TODO: fetch model from server
        # TODO: support permanent IDs
        username, name = name.split("/")
        return self.prepare_model({"username": username, "name": name})



# !/usr/bin/env python

from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# setup.py cannot safely import files, so read and exec to get the version
__version__ = None
exec(open("replicate/__about__.py").read())

setup(
    name="replicate",
    packages=["replicate"],
    version=__version__,
    description="Python client for Replicate",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Replicate, Inc.",
    license="BSD",
    url="https://github.com/replicate/replicate-python",
    python_requires=">=3.6",
    install_requires=["requests", "pydantic"],
    classifiers=[],
)
